/* prismstumbler.c - Utility Scan for 802_11 networks 
 * 
 * This utility is written for use with IEEE 802.11 adapters based
 * on Intersil's PRISM II chipset (PCMCIA). 
 * The linux driver for these cards can be found on www.linux-wlan.com
 * It has been verified with a D-LINK DWL-6540 WL100 IEEE 802.11 adapter. 
 * Created by: Jan Fernquist  Jan.B.Fernquist@telia.com
 * 
 *  The main idea for this program create a NetStumbler like utlity to find W-LANS
 *
 * Base code was from PrismDump.
 * Code also Snarfed from Ethreal and airsnort
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*
 * RevHist
 *  
 * When		Who		What
 * 02-03-06     jafer		New options -S N  Tune in on Single channel	
 *				-F  F  Save data to file F as well as stdout
 *			        Time Field has a decimal to get difference form packet arriving
 *			   	at the same second. Used in Perlprog to sort the last seen packet firts
 *
 * 02-01-09	jafer		Embedded use of the functtion wlanctl-ng by using 
 *				the actual program
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <asm/types.h>
#include <linux/netlink.h>
#include <netinet/in.h>


#define HELPTEXT "Call with no parameters or with the following options\n\
	-S Channel   - Only listen on a specific Channel (1-14)\n\
	-F FileName  - Save output to an file as well as stdout\n"

#define MAX_BUFFER_SIZE 4000	/* Size of receive buffer */
#define DEVNAME_LEN 16
#define __WLAN_ATTRIB_PACK__       __attribute__ ((packed))
#define MCAST_GRP_SNIFF 0x00000002	/* Value used to notify netlink  */



static unsigned int stop_sniffing = 0;
/*
 * Packet Types
 */
#define MGT_PROBE           0x0040       /* Client ProbeRequest        */
#define MGT_PROBE_RESP           0x0050       /* AP ProbeReponse        */
#define MGT_ASSOC_REQ        0x0000     /* Client Assocrequest */

#define MGT_BEACON           0x0080       /* Management - Beacon frame           */
#define FLAG_TO_DS              0x01
#define FLAG_FROM_DS            0x02
#define FLAG_WEP                0x10

#define IS_TO_DS(x)            ((x) & FLAG_TO_DS)
#define IS_FROM_DS(x)          ((x) & FLAG_FROM_DS)
#define IS_WEP(x)              ((x) & FLAG_WEP)

/* ************************************************************************* */
/*        Logical field codes (IEEE 802.11 encoding of tags)                 */
/* ************************************************************************* */
#define TAG_SSID           0x00
#define TAG_SUPP_RATES     0x01
#define TAG_FH_PARAMETER   0x02
#define TAG_DS_PARAMETER   0x03
#define TAG_CF_PARAMETER   0x04
#define TAG_TIM            0x05
#define TAG_IBSS_PARAMETER 0x06
#define TAG_CHALLENGE_TEXT 0x10


/*
 * How many packet should we read if we have a signal on a channel????
 * After that we force a channel change anyway
 */
#define CHANNEL_READ_COUNT 3

/*
 * The basic information types in the sniffed packed from prism
 */
typedef struct {
  __u32 did __attribute__ ((packed));
  __u16 status __attribute__ ((packed));
  __u16 len __attribute__ ((packed));
  __u32 data __attribute__ ((packed));
} p80211item_t;

/*
 * Just before the payload, the prismcard
 * puts in a lot of other info 
 * like signal quality an stuff
 */
typedef struct {
  __u32 msgcode __attribute__ ((packed));
  __u32 msglen __attribute__ ((packed));
  __u8 devname[DEVNAME_LEN] __attribute__ ((packed));
  p80211item_t hosttime __attribute__ ((packed));
  p80211item_t mactime __attribute__ ((packed));
  p80211item_t channel __attribute__ ((packed));
  p80211item_t rssi __attribute__ ((packed));
  p80211item_t sq __attribute__ ((packed));
  p80211item_t signal __attribute__ ((packed));
  p80211item_t noise __attribute__ ((packed));
  p80211item_t rate __attribute__ ((packed));
  p80211item_t istx __attribute__ ((packed));
  p80211item_t frmlen __attribute__ ((packed));
} AdmInfo_t;

/*
 * Overlay to directly index varius types of data
 * in Mgmt Frames
 */
typedef struct {
	unsigned short frametype;
	unsigned short duration;
	unsigned char DestAddr[6];
	unsigned char SrcAddr[6];
	unsigned char BssId[6];
	unsigned short FragSeq;
        unsigned short TimeStamp[4];
	unsigned short BeaconInterval;
	unsigned short Capabilities;
}FixedMgmt_t;

typedef struct {
	unsigned short frametype;
	unsigned short duration;
	unsigned char DestAddr[6];
	unsigned char SrcAddr[6];
	unsigned char BssId[6];
	unsigned short FragSeq;
}ProbeFixedMgmt_t;

typedef struct {
	char DestMac[20];
	char SrcMac[20];
	char BssId[20];
	char SSID[80];
	int hasWep;
	int isAp;
	int  Channel;
	int Signal;
        int Noise;
        int FrameType;
}ScanResult_t;

/*
 * Use globals, ugly but efficient
 */
static int RawSock;
static	 struct sockaddr_nl nl_sk_addr;
static ScanResult_t Res;
static int SingleChannel = 0;
static char * OutFileName = NULL;
static FILE * OutFile;
/*
 * Some forward refs
 */
int getPacket(unsigned char *buf,int maxlen,int timeout) ;
void closePacket(void);
int openPacket(void);
int selectChannel(int channel);
int processPacket(unsigned char *packet, int len,int RealChannel);
void shutCard(void);


/*
 * Signal Handler
 */
void stop_signal () {
  fprintf (stderr,"Received CTRL-C - sniffing aborted\n");
  stop_sniffing = 1;
}

/*
 * Help the poor guy!
 */
void HelpAndBye(void){
	fprintf(stderr,"%s",HELPTEXT);
	exit(1);
}

int main (int argc, char **argv) {
	int  n; 
	int recvlen;
	unsigned char msgbuf[MAX_BUFFER_SIZE];
	int Channel=0;
	int MaxPackPerChannel;
	struct  tm *t;
	time_t now,last_time;
	int Time_Add;
	int c;

	while(  ( c = getopt(argc,argv,"S:F:") ) != EOF ){
		switch(c){
			case 'S':   
					SingleChannel = atoi(optarg);
					if( SingleChannel < 1 || SingleChannel > 14 ){
							HelpAndBye();
					}
					
			case 'F':   	OutFileName = optarg;
					break;
			case '?':       HelpAndBye();
					break;
		}
	}

		
	if( OutFileName ){
		if(( OutFile = fopen(OutFileName,"w")) == NULL ){
			fprintf(stderr,"Cant open outfile %s for writing\n",OutFileName);
			exit(1);
		}
	}
		
		
	setpriority (PRIO_PROCESS, 0, -20);

	signal (SIGINT, stop_signal);
	fprintf (stderr, "Use CTRL-C to stop sniffing\n");



	if(  openPacket() < 0 ){
		perror(" Cant open Netlink Socket");
		exit(1);
	}  


	time(&last_time);
	 Time_Add = 0	;

	while( stop_sniffing == 0){

		if( ! SingleChannel ){
			Channel++;
			if( Channel > 14) Channel = 1;
		}else{
			Channel = SingleChannel;
		}

		if( selectChannel(Channel) != 0 ) break;
	     
		 MaxPackPerChannel = CHANNEL_READ_COUNT;
		do{
			memset (msgbuf, 0, MAX_BUFFER_SIZE);
			recvlen = getPacket(msgbuf, MAX_BUFFER_SIZE, 100);

			if( recvlen < 0 ) break;  /* Error */

			if( recvlen ==  0 ) break;  /* Nothing to read */

			if(  processPacket(msgbuf, recvlen,Channel)  ){  /* Got A valid inetersting ?*/

				if( Res.Channel < 1 || Res.Channel > 14 ){
						 continue; /* Bogus pkt*/
				}
				time(&now);
				if( now == last_time ){
					Time_Add++;
				}else{
					last_time = now;
					Time_Add = 0;
				}
						
				if( OutFile ){
					fprintf(OutFile,"%d.%.2d;%s;%d;%d;%d;%d;%d;%s;%s;%s;%.4X\n",
						now,Time_Add,
						Res.SSID,Res.Channel,Res.hasWep,Res.isAp,
						Res.Signal,Res.Noise,
						Res.DestMac,Res.SrcMac,Res.BssId ,Res.FrameType
					);	
				}
				printf("%d.%.2d;%s;%d;%d;%d;%d;%d;%s;%s;%s;%.4X\n",
					now,Time_Add,
					Res.SSID,Res.Channel,Res.hasWep,Res.isAp,
					Res.Signal,Res.Noise,
					Res.DestMac,Res.SrcMac,Res.BssId ,Res.FrameType
				);	
				fflush(stdout);
			}

		}while(  --MaxPackPerChannel > 0 );/* Until we read enough */
		

	}/*  while  still sniffing */

	if(OutFile){
		fclose(OutFile);
	}
	closePacket();
	shutCard();
	exit(0);
}

/*
 * Open the NETLINK Device
 */
int openPacket(void){

	if ((RawSock = socket (PF_NETLINK, SOCK_RAW, MCAST_GRP_SNIFF)) < 0) {
	      perror ("Could not create socket");
	      return -1; 
	}

	memset (&nl_sk_addr, 0, sizeof (struct sockaddr_nl));
	nl_sk_addr.nl_family = (sa_family_t) PF_NETLINK;
	nl_sk_addr.nl_pid = (unsigned int) getpid ();

	nl_sk_addr.nl_groups = MCAST_GRP_SNIFF;

	if (bind (RawSock, (struct sockaddr *) &nl_sk_addr, 
				sizeof (struct sockaddr_nl)) < 0) {
		perror ("Call to bind failed");
		return -1;
	}
	return RawSock;
}


/*
 * Just a wrapper....
 */
void closePacket(){
	close(RawSock);
}


/*
 * Check if data is avaliable 
 * Grab a packet from Interface
 */
int getPacket(unsigned char *buf,int maxlen,int timeout) {
	fd_set rs;
	int r;
	struct timeval tm;
	struct timeval *ptm;
	int  n, oi;

	FD_ZERO(&rs);
	FD_SET(RawSock,&rs);
	if (timeout>=0) {
		tm.tv_sec=timeout/1000;
		tm.tv_usec=(timeout%1000)*1000;
		ptm=&tm;
	} else {
		ptm=NULL;
	}
	r=select(RawSock+1,&rs,NULL,NULL,ptm);

	if (r<0){ 
		perror("Select failed");  
		return(-1);
	}
	if (r==0) {
		if (timeout>=0){ 
			/*printf("Timeout\n");*/  
			return(0);
		}
	}

	if (FD_ISSET(RawSock,&rs)) {
		r=recv(RawSock,buf,maxlen,0);

		if (r<0){ 
			printf(" recv failed\n"); 
			return(-1);
		}
	    return(r);
	  }

	  return(0); /* Noting to read */
}


#ifdef USE_SYSTEM
/*
 * This is a ugly hack just to set the channel to listen on
 */
void shutCard(void){
	char Cmd[255];

	sprintf(Cmd,"/sbin/wlanctl-ng wlan0 lnxreq_wlansniff enable=false > /dev/null");
	system(Cmd);
}

/*
 * This is a uglc hack just to set the channel to listen on
 */
int selectChannel(int channel){
	char Cmd[255];

	sprintf(Cmd,"/sbin/wlanctl-ng wlan0 lnxreq_wlansniff enable=false > /dev/null");
	system(Cmd);
	sprintf(Cmd,
	   "/sbin/wlanctl-ng wlan0 lnxreq_wlansniff channel=%d enable=true > /dev/null",
		channel);
	system(Cmd);
	return 0 ;
}
#else
void shutCard(void){
	char *argv[10];
	int argc;

	argv[0] = "wlanctl";
	argv[1] = "wlan0";
	argv[2] = "lnxreq_wlansniff";
	argv[3] = "enable=false";
	argc=4;
	wlanctl(argc,argv);

}
int selectChannel(int channel){
	char Channel[30];
	char *argv[10];
	int argc;
	sprintf(Channel,"channel=%d",channel);

	argv[0] = "wlanctl";
	argv[1] = "wlan0";
	argv[2] = "lnxreq_wlansniff";
	argv[3] = "enable=false";
	argc=4;
	wlanctl(argc,argv);
	argv[3] = Channel;
	argv[4] = "enable=true";
	argc=5;
	wlanctl(argc,argv);
}
	
#endif
/*
*
 * Get the different data from the MGMT header. Fill global struct
 */
int processPacket(unsigned char *packet, int len,int RealChannel){
	FixedMgmt_t  *M;
	ProbeFixedMgmt_t  *P;
	AdmInfo_t    *A; 
  	
	unsigned char *admbits,*fixbits,*varBits;
	int i;
	int tagType,tagLen;
	unsigned char TextBuff[256];

	admbits = packet;
	fixbits = &packet[sizeof(AdmInfo_t)]; /* This is where the payload starts*/

	A = (AdmInfo_t *)admbits; /* First the frame from the card */
	M =  (FixedMgmt_t *) fixbits; /*Then the actual data*/
	P =  (ProbeFixedMgmt_t *) fixbits; /*Then the actual data*/

	memset(&Res,0,sizeof(Res));

	Res.FrameType = M->frametype;
	

	if( M->frametype  == MGT_PROBE_RESP || M->frametype == MGT_BEACON ){ 
			Res.isAp   = IS_TO_DS(M->Capabilities) != 0;
			Res.hasWep = IS_WEP(M->Capabilities) != 0;

	}
	if( M->frametype  == MGT_PROBE || M->frametype == MGT_BEACON ){ 
			if(! ( M->DestAddr[0] == 0xff && M->DestAddr[1] == 0xff &&
				M->DestAddr[2] == 0xff && M->DestAddr[3] == 0xff &&
				M->DestAddr[4] == 0xff && M->DestAddr[5] == 0xff 
			     ))  return 0;
	}

	if( M->frametype  == MGT_PROBE  ){ 
		 Res.isAp   = 3  ;
		 Res.hasWep =  0;
		 Res.Channel = RealChannel ;
	}



	sprintf(Res.DestMac,"%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
		M->DestAddr[0], M->DestAddr[1],
		 M->DestAddr[2], M->DestAddr[3],
		 M->DestAddr[4], M->DestAddr[5]
	);
	sprintf(Res.SrcMac,"%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
		M->SrcAddr[0], M->SrcAddr[1],
		 M->SrcAddr[2], M->SrcAddr[3],
		 M->SrcAddr[4], M->SrcAddr[5]
	);

	sprintf(Res.BssId,"%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
		M->BssId[0], M->BssId[1],
		 M->BssId[2], M->BssId[3],
		 M->BssId[4], M->BssId[5]
	);

	Res.Signal = A->signal.data;
	Res.Noise = A->noise.data;


	if(  M->frametype == MGT_PROBE ){
		if( len <= sizeof(ProbeFixedMgmt_t) ) return 0;

		varBits = &fixbits[sizeof(ProbeFixedMgmt_t)];

	}else{ /* Beacon and Probe Response */
		if( len <= sizeof(FixedMgmt_t) ) return 0;

		varBits = &fixbits[sizeof(FixedMgmt_t)];
	}

	/* Get the tagged values 
	 * There is a type and a length field
	 * Followed by tha actual data 
         */
	while( varBits < (packet+len)){

		tagType = varBits[0]; 
		tagLen = varBits[1];
		varBits+=2;
	
	
		switch(tagType){
			case TAG_SSID:
					strncpy(Res.SSID,varBits,tagLen);
					break;

			case TAG_DS_PARAMETER:
						if(tagLen == 1){
							Res.Channel = *varBits;
						}
						break;
			/* Skip all other values for the moment */
			case TAG_SUPP_RATES:
			case TAG_FH_PARAMETER:
			case TAG_CF_PARAMETER:
			case TAG_TIM:
			case TAG_IBSS_PARAMETER :
			case TAG_CHALLENGE_TEXT:	
						 break;
		}	
		varBits+=tagLen;
	}
	return 1;
}


