#!/usr/bin/perl
############################################################
# This is a simple  HTML Output generator for  Prismstumbler
# Runs prismstumbler as a subprocess
# Listen on port 9000  for HTTP requests.
# Sends a simple Table of the data captured
#
#  RevHist
#  
#  When 	Who			What
#  02-03-06     jafer		      New column Packet type added
#				      Time data has a decimal to be able tos sort on time seend
#                                     Output reverse sorted on time first Seen
#				      I.E The Latest seen is always first
#				      Options Are processed and passed to C-program
#				      New Option -P which tells webserver to listen
#				      on different port
#
#  02-01-09	pupok(Janie Mehew)    Colorpatch to highlight active AP's
#  
#
#
#
################################################################
use IO::Handle;
use IO::File;
use IO::Socket::INET;
use IO::Select;
use Getopt::Std;


getopt('R:F:S:P:');

if( $opt_R ){
        $PrismStumbler = "./slowcat.pl  $opt_R " ;
}else{ 
	$PrismStumbler = "./prismstumbler ";

	if( $opt_S ){
		$PrismStumbler .= "-S $opt_S "; 
	}
	if( $opt_F ){
		$PrismStumbler .= "-F $opt_F "; 
	}
}

if( $opt_P ){
	$WebPort = $opt_P;
}else{
	$WebPort = 9000;
}



#
# This is where the  file that maps Ethernet prefixes
# to Nmanufacturers name resides
$EthersFile="./manufacturers.dat";

#
# Build the Ethernet-Manufacurer map
#
%EtherList = &ReadEthers($EthersFile);

if( ! defined( $EtherList{"00:00:00"} ) ) {
	print "Cant read  Manufacurerlist $EthersFile\n";
	exit(1);
}


#########
# The Header that the simple HTTP-server always will send
#
$Header = <<ENDOFHEADER;
HTTP/1.1 200 OK\r
Server: HomeBrew/5.0\r
Content-Type: text/html\r
\r\n\r
ENDOFHEADER

#
# The fixed part of the content that the webserver will send
#
$Cont = <<ENDOFHEADER;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">
<html><head><title>PrismStumbler</title>
<META HTTP-EQUIV="Refresh" CONTENT="4; URL=http://localhost:9000/">
</head><body bgcolor="#FFFFFF">
ENDOFHEADER

%FoundArr;  
%FirstHash;  


$WebServer = IO::Socket::INET->new( Listen => 10,
				    LocalAddr => 'localhost',
				    LocalPort => $WebPort,
				    Proto => 'tcp' );

die "Cant listen on 9000" if( ! defined $WebServer);

$Sniffer = new IO::File "$PrismStumbler |" ;

die "Cant open $PrismStumbler" if( ! defined $Sniffer);

###############
# Main Loop 

$Sel = IO::Select->new();
$Sel->add($WebServer);
$Sel->add($Sniffer);

while(1){

	@Ready = $Sel->can_read(1);

	foreach $fh ( @Ready ){
		if( $fh == $Sniffer ){	
			&ReadFromSniffer(); # Got SniffData 
		}
		if( $fh == $WebServer ){ #WebClient
			$new = $WebServer->accept();
			&GetWebRequest($new);
			&SendWebData($new);
			$new->close();
		}
	}
}



#
# Read a line form the sniffer
# Put the result in the FoundArray.
# Update some values should it already exist
#
sub ReadFromSniffer{
	my ($RealTS,$TS,$SSID,$Channel,$Wep,$Ap,$Signal,$Nose,$Dest,$Source,$BSSID, $PacketType); 

	$_ = <$Sniffer>;
	chop;
	($RealTS,$SSID,$Channel,$Wep,$Ap,$Signal,$Nose,$Dest,$Source,$BSSID,$PacketType) = split(';');

	#return if(  $Ap != 1 ); # We just care about AP's
	$TS = $RealTS;
	$TS =~ s/\.*$//;  # Remove the decimal part

	if(defined( $FoundArr{$BSSID})  ){
		# Update some Fields....
		$FoundArr{$BSSID}{LastSeen} = &TS($TS); 
		$FoundArr{$BSSID}{SeenTS} = time;
		$FoundArr{$BSSID}{Signal} = $Signal; 
		$FoundArr{$BSSID}{MaxSignal} = 
			$Signal if( $Signal > $FoundArr{$BSSID}{MaxSignal}); 

		$FoundArr{$BSSID}{MinSignal} = 
			$Signal if( $Signal < $FoundArr{$BSSID}{MinSignal}); 
		
		$FoundArr{$BSSID}{Noise} = $Nose; 
		$FoundArr{$BSSID}{PacketType} = $PacketType; 
		$FoundArr{$BSSID}{SSID} = $SSID;  # If we got  a probe resp this could change
		$FoundArr{$BSSID}{FoundPack} += 1 ; 
	}else{
	      # New entry. Update all
		$FirstHash{$RealTS} = $BSSID;
		$FoundArr{$BSSID}{LastSeen} = &TS($TS); 
		$FoundArr{$BSSID}{SeenTS} = time;
		$FoundArr{$BSSID}{FirstSeen} = &TS($TS); 
		$FoundArr{$BSSID}{Signal} = $Signal; 
		$FoundArr{$BSSID}{MaxSignal} = $Signal; 
		$FoundArr{$BSSID}{MinSignal} = $Signal; 
		$FoundArr{$BSSID}{Noise} = $Nose; 
		$FoundArr{$BSSID}{SSID} = $SSID; 
		$FoundArr{$BSSID}{Channel} = $Channel; 
		$FoundArr{$BSSID}{WEP} = $Wep; 
		$FoundArr{$BSSID}{AP} = $Ap; 
		$FoundArr{$BSSID}{Dest} = $Dest; 
		$FoundArr{$BSSID}{Source} = $Source; 
		$FoundArr{$BSSID}{BSSID} = $BSSID; 
		$FoundArr{$BSSID}{FoundPack} = 1;  
		$FoundArr{$BSSID}{PacketType} = $PacketType; 
		$FoundArr{$BSSID}{Manuf} = &FindManuf($BSSID) ;  
	}	
}


#########
# Get the HTTP-Headers, Ignore what it says
#
sub GetWebRequest{ 

	my( $handle ) =@_;

	while(<$handle>){
		chop;
		s/\r//;
		#print "$_\n";
		if(length($_) < 1 ){ 
			last;
		}
	}
}


	
#############
# Build content from arrays
#
sub SendWebData{

	my( $handle ) =@_;

	my $RetVal;
	my $Key;	
	my $bgcolor;
	my($Fs,$Ls,$Ns,$Mf);

	printf  $handle $Header; 
	printf  $handle $Cont; 

	print $handle  "<table border=\"0\" cellspacing=\"3\" cellpadding=\"2\"><tr>";


	foreach $Key  ( "No pkt","FirstSeen","LastSeen","SSID","Channel","Wep","Ap","BSSID","Manuf","Signal",
			"MaxSig","MinSig","PacketType"
		){

			print $handle "<td><strong>$Key</strong></td>";

	}
	print $handle  "</tr>\n";
	# foreach  $Key ( keys %FoundArr ){
	foreach  $Val ( reverse sort keys %FirstHash  ){
		$Key = $FirstHash{$Val};
		next if length("$Val") < 6;
		#print "Found Key :$Key: Val :$Val:\n";
		$bgcolor = &getColor($FoundArr{$Key}{'SeenTS'});
		print $handle "<tr bgcolor=$bgcolor>";
		foreach $Value 	( 'FoundPack','FirstSeen','LastSeen','SSID',
 				  'Channel','WEP','AP','BSSID',
			     	  'Manuf','Signal','MaxSignal','MinSignal','PacketType'
				){
				print $handle 
			"<td><center>$FoundArr{$Key}{$Value}</center></td>\n";
		
		}
		print $handle "</tr>"
	}
	print $handle "</table></body></html>"	
}

#
# Make a HH:MM:SS Text timestamp
#
sub TS{
	my($Now)=@_;

	my($SEC,$MIN,$HOUR);

	my($Ret);

	($SEC,$MIN,$HOUR)=localtime($Now);
	$Ret = sprintf("%.2d:%.2D:%.2d",$HOUR,$MIN,$SEC);
	$Ret;
}

#
# return a color based on time since last seen.
#

sub getColor{
	my($seen)=@_;
	my($passed);
	my($Ret);
	my($color);


	$passed = time - $seen;
	if ( $passed > 30 ) {
		$Ret = sprintf("#%x%x%x",255,255,255);
	}
	else {
		$color = 8 * $passed;
		$Ret = sprintf("#%x%x%x",255,$color,255);
	}
	$Ret;
}

#
# Find the manufacturer given the MAC address
#
sub FindManuf{
	my( $Ether ) = @_;
	#print "Loking for $Ether\n";
	$_ = $Ether;
	if( /([0-9,A-F][0-9,A-F]:[0-9,A-F][0-9,A-F]:[0-9,A-F][0-9,A-F])(.*$)/ ){
	#	printf "Found $1 = %s\n",$EtherList{$1};
		return $EtherList{$1};
	}
}

#
# Build the manufacturer list
#
sub ReadEthers{
	my ($File) = @_;
	my($Fh);
	my(%Ret);
	$Fh = new IO::File "<$File";
	
	while(<$Fh>){
		chop;
		next if /#/;
		next if length($_) <5;

		if( /^([0-9,A-F][0-9,A-F]:[0-9,A-F][0-9,A-F]:[0-9,A-F][0-9,A-F])\s(.*$)/ ){
			$Ret{$1}=substr($2,0,12);
		}
	}	
	$Fh->close();
	%Ret;
}
	
